# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
from .callbacks import Callback, Callbacks, Cbs
import k1lib; from k1lib import squeeze; import k1lib.cli as cli
from functools import partial; plt = k1lib.dep("matplotlib.pyplot")
from typing import List, Tuple, Dict, Iterator, Union, Any, Callable
try: import torch; import torch.nn as nn; hasTorch = True
except:
    torch = k1lib.Object().withAutoDeclare(lambda: type("RandomClass", (object, ), {}))
    nn = k1lib.Object().withAutoDeclare(lambda: type("RandomClass", (object, ), {})); hasTorch = False
__all__ = ["HookModule"]
class Handles:                                                                   # Handles
    def __init__(self):                                                          # Handles
        self.forward = None; self.backward = None                                # Handles
    def remove(self):                                                            # Handles
        if self.active:                                                          # Handles
            self.forward.remove(); self.forward = None                           # Handles
            self.backward.remove(); self.backward = None                         # Handles
    @property                                                                    # Handles
    def active(self):                                                            # Handles
        if self.forward != None and self.backward != None: return True           # Handles
        elif self.forward == None and self.backward == None: return False        # Handles
        raise Exception("Supposed to be unreachable")                            # Handles
class Data(k1lib.Object):                                                        # Data
    def __init__(self):                                                          # Data
        super().__init__(); self.withAutoDeclare(lambda: [])                     # Data
class ModuleData:                                                                # ModuleData
    def __init__(self): self.forward = Data(); self.backward = Data()            # ModuleData
    def _plot(self, axes, field:str, rangeSlice:slice, f):                       # ModuleData
        forwardData = self.forward[field]; step = rangeSlice.step or 1           # ModuleData
        backwardData = self.backward[field]                                      # ModuleData
        if len(forwardData) == 0 or len(backwardData) == 0: return               # ModuleData
        fR, bR = k1lib.Range.proportionalSlice(len(forwardData), len(backwardData), rangeSlice) # ModuleData
        axes[0].plot(fR.range_[::step] | f | cli.deref(), forwardData[fR.slice_][::step] | f | cli.deref(), alpha=0.5) # ModuleData
        axes[1].plot(bR.range_[::step] | f | cli.deref(), backwardData[bR.slice_][::step] | f | cli.deref(), alpha=0.5) # ModuleData
    def __repr__(self):                                                          # ModuleData
        return """Module's saved data. can...
- d.forward: to get data stored during forward pass
- d.backward: to get data stored during backward pass"""                         # ModuleData
_Fn = Callable[[Data, nn.Module, Tuple[torch.Tensor], Tuple[torch.Tensor]], None] # ModuleData
class Function:                                                                  # Function
    def __init__(self, f:_Fn, name=None):                                        # Function
        self.f = f; self.name = name or "f(<no name>)"                           # Function
    def __call__(self, *args, **kwargs):                                         # Function
        self.f(*args, **kwargs)                                                  # Function
def hook(every, fns:List[Function], *args):                                      # hook
    if every.value: [fn(*args) for fn in fns]                                    # hook
class Module:                                                                    # Module
    def __init__(self, module:nn.Module):                                        # Module
        self.nn = module                                                         # Module
        self.handles = Handles()                                                 # Module
        self.data = ModuleData()                                                 # Module
        self.name = module.__class__.__name__                                    # Module
    def registerHooks(self, forwardFns:List[Function], backwardFns:List[Function], every): # Module
        self.handles.forward = self.nn.register_forward_hook(partial(hook, every, forwardFns, self.data.forward)) # Module
        self.handles.backward = self.nn.register_full_backward_hook(partial(hook, every, backwardFns, self.data.backward)) # Module
        return self                                                              # Module
    def unregisterHooks(self): self.handles.remove()                             # Module
    def __repr__(self):                                                          # Module
        return f"""Module `{self.name}`. Use...
- m.data: to get data stored
- m.nn: to get actual nn.Module object
- m.plot("means", "stds"): to plot simple statistics"""                          # Module
@k1lib.patch(Cbs)                                                                # Module
class HookModule(Callback):                                                      # HookModule
    """Hooks into selected modules in the network, and
execute functions like .mean(), .std(). This is fairly
complicated, and I highly recommend displaying this
callback in a cell for more info"""                                              # HookModule
    def __init__(self, persistent:bool=False):                                   # HookModule
        """
:param persistent: whether to save results across
    runs. If false, then can execute `.reset()` to
    reset everything"""                                                          # HookModule
        super(HookModule, self).__init__()                                       # HookModule
        self.modules:List[Module] = []                                           # HookModule
        self.forwardFns:List[Function] = []                                      # HookModule
        self.backwardFns:List[Function] = []                                     # HookModule
        self.cleanFns = []; self.persistent = persistent                         # HookModule
        self.every = k1lib.Every(3)                                              # HookModule
    def reset(self):                                                             # HookModule
        """Intended to be called by end user only, to reset
        everything if choose to persist results across runs."""                  # HookModule
        self._end(); self._start()                                               # HookModule
    def persist(self):                                                           # HookModule
        """By default, data will be erased and populated on each run. If
you want the data to persist across runs, call this."""                          # HookModule
        self.persistent = True                                                   # HookModule
    def startRun(self):                                                          # HookModule
        if (not self.persistent) or (len(self.modules) == 0): self._start()      # HookModule
    def startBatch(self): self.every()                                           # HookModule
    def _registerHooks(self):                                                    # HookModule
        for module in self.modules:                                              # HookModule
            module.registerHooks(self.forwardFns, self.backwardFns, self.every)  # HookModule
    def _unregisterHooks(self):                                                  # HookModule
        for module in self.modules: module.unregisterHooks()                     # HookModule
    def endRun(self):                                                            # HookModule
        if not self.persistent: self._end()                                      # HookModule
    def suspend(self):                                                           # HookModule
        self.actuallyRestore = len(self) == 0 or self[0].handles.active          # HookModule
        if self.actuallyRestore: self._unregisterHooks()                         # HookModule
    def restore(self):                                                           # HookModule
        if self.actuallyRestore:                                                 # HookModule
            self._registerHooks()                                                # HookModule
            self.actuallyRestore = False                                         # HookModule
    def __getitem__(self, idx):                                                  # HookModule
        if type(idx) == int: return self.modules[idx]                            # HookModule
        answer = HookModule(self.persistent)                                     # HookModule
        answer.modules = self.modules[idx]                                       # HookModule
        return answer                                                            # HookModule
    def __len__(self): return len(self.modules)                                  # HookModule
    def __repr__(self):                                                          # HookModule
        f = '\n'.join([f'  - {fn.name or str(fn)}' for fn in self.forwardFns])   # HookModule
        f = "" if f == "" else f"Forward hooks:\n{f}\n"                          # HookModule
        b = '\n'.join([f'  - {fn.name or str(fn)}' for fn in self.backwardFns])  # HookModule
        b = "" if b == "" else f"Backward hooks:\n{b}\n"                         # HookModule
        n = '\n'.join([f'  {i}. {data.name}' for i, data in enumerate(self)])    # HookModule
        excludes = {"withForwardHook", "withBackwardHook", "withHook", "withCheckpoint"} # HookModule
        withs = '\n'.join([f"- m.{key}()" for key in dir(self) if key.startswith("with") and key not in excludes]) # HookModule
        return f"""{super()._reprHead} with {len(self)} modules:\n{n}\n{f}{b}
Use...
- m.plot("means", "stds"): to plot simple statistics
- m[i]: to get a specific module
- m[a:b]: to get a new HookModule with selected modules
- m.css("..."): to select a specific subset of modules only
- m.withHook(hookCb): to hook a specific callback function
- m.clearHooks(): to clear all hooks
{super()._reprCan}

Built-in `with-` functions:\n{withs}"""                                          # HookModule
@k1lib.patch(HookModule)                                                         # HookModule
def _start(self):                                                                # _start
    self.modules = []                                                            # _start
    for nn, sel in zip(self.l.model.modules(), self.l.selector.modules()):       # _start
        if "HookModule" in sel: self.modules.append(Module(nn))                  # _start
    self._registerHooks()                                                        # _start
@k1lib.patch(HookModule)                                                         # _start
def _end(self):                                                                  # _end
    for module in self.modules:                                                  # _end
        for cleanFn in self.cleanFns:                                            # _end
            cleanFn(module.data)                                                 # _end
    self._unregisterHooks()                                                      # _end
@k1lib.patch(HookModule)                                                         # _end
def withForwardHook(self, hook:_Fn, name:str=None):                              # withForwardHook
    """Adds a hook to the forward pass. See :func:`~k1lib.callbacks.hookModule.HookModule.withHook`""" # withForwardHook
    self.forwardFns += [Function(hook, name)]; return self                       # withForwardHook
@k1lib.patch(HookModule)                                                         # withForwardHook
def withBackwardHook(self, hook:_Fn, name:str=None):                             # withBackwardHook
    """Adds a hook to the backward pass. See :func:`~k1lib.callbacks.hookModule.HookModule.withHook`""" # withBackwardHook
    self.backwardFns += [Function(hook, name)]; return self                      # withBackwardHook
@k1lib.patch(HookModule)                                                         # withBackwardHook
def withHook(self, hook:_Fn, name:str=None):                                     # withHook
    """Adds a hook to both the forward and backward pass.

:param hook: this function is expected to take in these parameters: **(data, module, inp, out)**

    :data: the injected dependency for you to store stuff.
        Initially, `data.max` is an empty list, and you can
        append to it directly, like this::

            data.max.append() # okay

        Later on, you can do things like::

            HookModule[i].forward.max

        and get the data you saved from the hook.
    :module: the module this function hooks into. Please
        refer to :func:`torch.nn.Module.register_forward_hook()` to
        know more.
    :inp: input (or grad of input) to the module
    :out: output (or grad of output) to the module
:param name: custom name for the function for nice displaying

See also: m.withForwardHook(), m.withBackwardHook()"""                           # withHook
    return self.withForwardHook(hook, name).withBackwardHook(hook, name)         # withHook
@k1lib.patch(HookModule)                                                         # withHook
def clearHooks(self):                                                            # clearHooks
    self._unregisterHooks()                                                      # clearHooks
    self.forwardFns = []; self.backwardFns = []                                  # clearHooks
    self.cleanFns = []; return self                                              # clearHooks
def meanCb(data, m, inp, out):                                                   # meanCb
    data.means.append(squeeze(out, hard=True).data.mean().item())                # meanCb
@k1lib.patch(HookModule)                                                         # meanCb
def withMeanRecorder(self):                                                      # withMeanRecorder
    """Records mean"""                                                           # withMeanRecorder
    return self.withHook(meanCb, "mean")                                         # withMeanRecorder
def stdCb(data, m, inp, out):                                                    # stdCb
    data.stds.append(squeeze(out, hard=True).data.std().item())                  # stdCb
@k1lib.patch(HookModule)                                                         # stdCb
def withStdRecorder(self):                                                       # withStdRecorder
    """Records standard deviation"""                                             # withStdRecorder
    return self.withHook(stdCb, "std")                                           # withStdRecorder
def minCb(data, m, inp, out):                                                    # minCb
    data.mins.append(squeeze(out, hard=True).data.min().item())                  # minCb
@k1lib.patch(HookModule)                                                         # minCb
def withMinRecorder(self):                                                       # withMinRecorder
    """Records min"""                                                            # withMinRecorder
    return self.withHook(minCb, "min")                                           # withMinRecorder
def maxCb(data, m, inp, out):                                                    # maxCb
    data.maxs.append(squeeze(out, hard=True).data.max().item())                  # maxCb
@k1lib.patch(HookModule)                                                         # maxCb
def withMaxRecorder(self):                                                       # withMaxRecorder
    """Records max"""                                                            # withMaxRecorder
    return self.withHook(maxCb, "max")                                           # withMaxRecorder
@k1lib.patch(HookModule)                                                         # withMaxRecorder
def css(self, css:str):                                                          # css
    answer = HookModule()                                                        # css
    selector = k1lib.selector.select(self.l.model, css)                          # css
    d = {m.nn: m for m in self.modules}                                          # css
    for sel in selector.modules():                                               # css
        if "HookModule" in sel and sel.nn in d:                                  # css
            answer.modules.append(d[sel.nn])                                     # css
    return answer                                                                # css
def plotF(modules:HookModule, fields:List[str], f, rangeSlice:slice):            # plotF
    fig, axes = plt.subplots(len(fields), 2, figsize=(10, 3*len(fields)), dpi=100) # plotF
    axes = axes.reshape((-1, 2))                                                 # plotF
    for axs, field in zip(axes, fields):                                         # plotF
        for module in modules:                                                   # plotF
            module.data._plot(axs, field, rangeSlice, f)                         # plotF
        axs[0].set_title(f"Forward {field}")                                     # plotF
        axs[1].set_title(f"Backward {field}")                                    # plotF
    plt.figlegend([f"{i}. {module.name}" for i, module in enumerate(modules)], loc='center right') # plotF
@k1lib.patch(HookModule)                                                         # plotF
@k1lib.patch(Module)                                                             # plotF
def plot(self, *fields:List[str], f=cli.iden()):                                 # plot
    """Plots every simple (1 number saved/pass/module) fields.

:param fields: list of fields to plot. If none, then
    will automatically find all simple fields"""                                 # plot
    modules = [self] if isinstance(self, Module) else self                       # plot
    if len(modules) == 0: raise Exception("No modules to plot!")                 # plot
    if len(fields) == 0:                                                         # plot
        fields = []; forwardData = modules[0].data.forward                       # plot
        for field in forwardData.state.keys():                                   # plot
            if field.startswith("_"): continue                                   # plot
            fieldData = forwardData[field]                                       # plot
            if type(fieldData) == list and k1lib.isNumeric(fieldData[0]):        # plot
                fields.append(field)                                             # plot
    return k1lib.viz.SliceablePlot(partial(plotF, modules, fields, f))           # plot



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
"""
I have several machine learning tools running on my own cluster that's hosted
on https://mlexps.com/#kapi, and this module contains functions, classes and
clis that will contact that service. This is so that if I want to use a language
model in multiple notebooks, I'd have to load the model into my GPU for each
notebook, which would waste a lot of resources. I can't run a lot of notebooks
at the same time as I'd just run out of VRAM. So, by having dedicated services/demos,
I can really focus on serving things well and make it performant. For example::

    "some text"           | kapi.embed()    # returns embedding numpy array
    "What is Python? "    | kapi.complete() # returns string, completes the sentence
    "image.png" | toImg() | kapi.ocr()      # returns `Ocr` object, with bounding boxes and text content of all possible texts
    "cute anime girl"     | kapi.txt2im()   # generates an image from some description
    "image.png" | toImg() | caption()       # generates a caption of an image


"""
__all__ = ["status", "segment", "demo", "embed", "embeds", "complete",
           "ocr", "Ocr", "OcrBox", "tess",
           "txt2im", "caption", "speech", "summarize", "post"]
from k1lib.cli.init import BaseCli; import k1lib.cli.init as init
import k1lib.cli as cli, k1lib, base64, html, json
requests = k1lib.dep("requests"); k1 = k1lib
settings = k1lib.settings.cli
s = k1lib.Settings(); settings.add("kapi", s, "cli.kapi settings")
s.add("local", False, "whether to use local url instead of remote url. This only has relevance to me though, as the services are running on localhost")
def get(idx:str, json):                                                          # get
    """Sends a request to any service/demo on https://mlexps.com.
Example::

    # returns "13.0"
    kapi.get("demos/1-arith", {"a": 1, "b": 3, "c": True, "d": 2.5, "e": 10})

:param idx: index of the service, like "kapi/1-embed" """                        # get
    url = "http://localhost:9000" if s.local else "https://local.mlexps.com"     # get
    res = requests.post(f"{url}/routeServer/{idx.replace('/', '_')}", json=json).json() # get
    if not res["success"]: raise Exception(res["reason"])                        # get
    return res["data"]                                                           # get
def jsF_get(idx, dataIdx):                                                       # jsF_get
    url = "https://local.mlexps.com"                                             # jsF_get
    return f"""await (await fetch("{url}/routeServer/{idx.replace('/', '_')}", {{method: "POST", headers: {{ "Content-Type": "application/json" }}, body: JSON.stringify({dataIdx})}})).json()""" # jsF_get
def status():                                                                    # status
    """Displays a table of whether the services are online and available or not""" # status
    ["kapi/1-embed", "kapi/2-complete", "kapi/3-ocr", "kapi/4-txt2im", "kapi/5-caption", "kapi/6-speech"] | cli.apply(lambda x: [x, requests.get(f"https://local.mlexps.com/routeServer/{x.replace(*'/_')}/healthCheck").text == "ok"]) | cli.insert(["Service", "Online"]) | cli.display(None) # status
class segment(BaseCli):                                                          # segment
    def __init__(self, limit:int=2000):                                          # segment
        """Segments the input string by sentences, such that each segment's length is lower than the specified limit.
Example::

    # returns ['some. Really', 'Long. String', 'Just. Monika']
    "some. Really. Long. String. Just. Monika" | segment(15)

So, this will split the input string by ". ", then incrementally joins the strings together into segments.
This is useful in breaking up text so that it fits within language model's context size""" # segment
        self.limit = limit                                                       # segment
    def __ror__(self, text):                                                     # segment
        if not isinstance(text, str): raise Exception("Input is not a string!")  # segment
        data = [[]]; c = 0; limit = self.limit                                   # segment
        for line in text.split(". "):                                            # segment
            if c + len(line) > limit and c > 0: # if even a single sentence is too big, then just have a segment as that sentence, and don't push it to the next one # segment
                data.append([]); c = 0                                           # segment
            data[-1].append(line); c += len(line)+2                              # segment
        return data | cli.join(". ").all() | cli.deref()                         # segment
metas = {} # Dict[prefix -> demo meta]                                           # segment
class demo(BaseCli):                                                             # demo
    def __init__(self, prefix:str="demos_1-arith"):                              # demo
        """Sends a request to one of mlexps.com demos.
Example::

    # returns 21.0
    {"a": 3} | kapi.demo("demos/1-arith")
    # builds js interface that displays 21.0
    {"a": 3} | (toJsFunc() | kapi.demo("demos/1-arith")) | op().interface("jsone")
    # same as above, but the dictionary is formed in JS instead of Python
    3 | (toJsFunc() | aS("{'a': x}") | demo("demos/1-arith")) | op().interface("jsone")

You don't have to specify all params, just the ones you want to deviate from the defaults""" # demo
        prefix = prefix.replace(*"/_"); self.prefix = prefix                     # demo
        res = requests.get(f"https://mlexps.com/{prefix.replace(*'_/')}/demo_meta.json") # demo
        if not res.ok: raise Exception(f"Demo {prefix.replace(*'_/')} not found!") # demo
        if prefix not in metas: metas[prefix] = json.loads(res.text)             # demo
    def __ror__(self, d):                                                        # demo
        prefix = self.prefix; meta = metas[prefix]; kw = {}                      # demo
        for arg in meta["args"]:                                                 # demo
            a = meta["defaults"][arg]; anno = meta["annos"][arg]                 # demo
            if anno in ("checkbox", "bytes", "image", "serialized"): a = a       # demo
            elif anno == "dropdown": a = a[1][a[0]]                              # demo
            elif anno == "apiKey": a = k1lib.apiKey if hasattr(k1lib, "apiKey") else a[0] # demo
            else: a = a[0]                                                       # demo
            kw[arg] = k1lib.serve.webToPy(a, anno)                               # demo
        for k, v in d.items(): kw[k] = v                                         # demo
        for k, v in kw.items(): kw[k] = k1lib.serve.pyToWeb(v, meta["annos"][k]) # demo
        url = "http://localhost:9003" if k1lib.settings.cli.kapi.local else "https://local.mlexps.com" # demo
        res = requests.post(f"{url}/routeServer/{prefix}", json=kw)              # demo
        if not res.ok: raise Exception(res.reason)                               # demo
        res = res.json()                                                         # demo
        if res["success"]: return k1lib.serve.webToPy(res["data"], meta["annos"]["return"]) # demo
        else: raise Exception(res["reason"])                                     # demo
    def __repr__(self): return f"<demo prefix='{self.prefix}'>"                  # demo
    def _repr_html_(self): s = html.escape(f"{self}"); return f"{s}{metas[self.prefix]['mainDoc']}" # demo
    def _jsF(self, meta):                                                        # demo
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto(); prefix = self.prefix; prefixSlash = prefix.replace(*"_/") # demo
        try: apiKey = k1lib.apiKey                                               # demo
        except: apiKey = ''                                                      # demo
        return f"""\
const {fIdx} = async ({dataIdx}) => {{
    const meta = await (await fetch("https://mlexps.com/{prefixSlash}/demo_meta.json")).json();
    const kw = {{}};
    for (const arg of meta.args) {{
        let a = meta.defaults[arg]; anno = meta.annos[arg];
        if (["checkbox", "image"].includes(anno)) a = a;
        else if (anno === "dropdown") a = a[1][a[0]];
        else if (anno === "apiKey") a = "{apiKey}";
        else a = a[0];
        kw[arg] = a
    }}
    for (const [k,v] of Object.entries({dataIdx})) {{ kw[k] = v; }}
    let res = await (await fetch("https://local.mlexps.com/routeServer/{prefix}", {{ method: "POST", headers: {{ "Content-Type": "application/json" }}, "body": JSON.stringify(kw) }})).text();
    try {{ res = JSON.parse(res); }} catch (e) {{ throw new Error(`Can't decode json of '${{res}}'`); }}
    if (!res.success) throw new Error(`Request failed: '${{res.reason}}'`);
    let data = res.data;
    if (meta.annos.return === "html") return atob(data);
    if (meta.annos.return === "image") return `<img src="data:image;base64,${{data}}">`
    return data;
}}""", fIdx                                                                      # demo
class embed(BaseCli):                                                            # embed
    def __init__(self):                                                          # embed
        """Gets an embedding vector for every sentence piped into this using `all-MiniLM-L6-v2`.
Example::

    # returns (384,)
    "abc" | kapi.embed() | shape()
    # returns (2, 384)
    ["abc", "def"] | kapi.embed().all() | shape()

- VRAM: 440MB
- Throughput: 512/s

See also: :class:`~k1lib.cli.models.embed`"""                                    # embed
        pass                                                                     # embed
    def __ror__(self, it): return self._all_opt([it]) | cli.item()               # embed
    def _all_opt(self, it:list[str]):                                            # embed
        for b in it | cli.batched(1024, True):                                   # embed
            yield from get("kapi/1-embed", {"lines": k1lib.encode(b)}) | cli.aS(k1lib.decode) # embed
class embeds(BaseCli):                                                           # embeds
    def __init__(self):                                                          # embeds
        """Breaks up some text and grab the embedding vectors of each segment.
Example::

    "sone long text" | kapi.embeds() # returns list of (segment, numpy vector)

This is just a convenience cli. Internally, this splits the text up using :class:`segment`
and then embeds each segment using :class:`embed`
"""                                                                              # embeds
        pass                                                                     # embeds
    def __ror__(self, it): return self._all_opt([it]) | cli.item()               # embeds
    def _all_opt(self, it:list[str]): return it | cli.apply(segment(700) | cli.iden() & embed().all() | cli.transpose()) | cli.deref() # embeds
class complete(BaseCli):                                                         # complete
    def __init__(self, prompt:str=None, maxTokens:int=200):                      # complete
        """Generates text from predefined prompts using `Llama 2`.
Example::

    # returns string completion
    "What is Python?" | kapi.complete()
    # returns list of string completions
    ["What is Python?", "What is C++?"] | kapi.complete().all()
    # returns list of string completions. The prompts sent to the server are ["<paragraph 1>\\n\\n\\nPlease summarize the above paragraph", ...]
    ["<paragraph 1>", "<paragraph 2>"] | kapi.complete("Please summarize the above paragraph").all()

- VRAM: 22GB
- Throughput: 8/s

:param max_tokens: maximum amount of tokens

See :class:`~k1lib.cli.models.complete`. That one is an older version using Google Flan T5 instead of llama 2""" # complete
        self.prompt = prompt; self.maxTokens = maxTokens                         # complete
    def __ror__(self, it): return self._all_opt([it]) | cli.item()               # complete
    def _all_opt(self, it:list[str]):                                            # complete
        if self.prompt: it = it | cli.apply(lambda x: f"{x}\n\n\n{self.prompt}: ") | cli.deref() # complete
        if not (isinstance(it, (list, tuple)) and isinstance(it[0], str)):       # complete
            raise Exception("You might have forgot to use .all(), like ['str1', 'str2'] | kapi.complete().all()") # complete
        it = it | cli.apply(lambda x: [x, self.maxTokens]) | cli.deref()         # complete
        return get("kapi/2-complete", {"prompts": json.dumps(it)}) | cli.aS(json.loads) # complete
    def _jsF(self, meta):                                                        # complete
        fIdx = cli.init._jsFAuto(); dataIdx = cli.init._jsDAuto()                # complete
        body = f"{{ prompts: JSON.stringify([{dataIdx}].map((x) => [`${{x}}\\n\\n\\n{self.prompt or ''}`, {cli.kjs.v(self.maxTokens)}])) }}" # complete
        return f"""
const {fIdx} = async ({dataIdx}) => {{
    const res = {jsF_get('kapi/2-complete', body)}
    return res[0]
}}""", fIdx                                                                      # complete
tf = k1.dep("torchvision.transforms")                                            # complete
class ocr(BaseCli):                                                              # ocr
    def __init__(self, paragraph:bool=False, resize=True, bs:int=10):            # ocr
        """Do OCR (optical character recognition) on some image.
Example::

    o = "some_image.png" | toImg() | kapi.ocr() # loads image and do OCR on them
    o     # run this in a separate notebook cell for an overview of where the boxes are
    o.res # see raw results received from the OCR service

That returns something like this::

    [[[771, 5, 813, 17], 'round', 0.7996242908503107],
     [[58, 10, 100, 34], '150', 0.883547306060791],
     [[166, 8, 234, 34], '51,340', 0.9991665158446097],
     [[782, 14, 814, 38], '83', 0.9999995785315409],
     [[879, 13, 963, 33], 'UPGRADes', 0.7625563055298393],
     [[881, 53, 963, 69], 'Monkey Ace', 0.9171751588707685],
     [[933, 133, 971, 149], '5350', 0.9001984000205994],
     [[873, 203, 911, 219], '5325', 0.481669545173645],
     [[931, 203, 971, 219], '5500', 0.7656491994857788],
     [[869, 271, 913, 291], 'G800', 0.31933730840682983],
     [[925, 271, 977, 291], '64600', 0.14578145924474253],
     [[871, 341, 911, 361], '5750', 0.5966295003890991],
     [[929, 341, 971, 361], '5850', 0.9974847435951233]]

First column is the bounding box (x1, y1, x2, y2), second column is the text,
and third column is the confidence, from 0 to 1.

Internally, this uses EasyOCR for the recognition. However, from my experience,
this doesn't always get it right. It's particularly bad at symbols like dollar
signs (it thinks it's "S", or "5" instead), periods or commads. So, you can refine
each of the bounding boxes like this::

    ocr = someImg | kapi.ocr()
    ocr[4] | toImg() | kapi.tess() # returns string, uses tesseract OCR instead of EasyOCR for more accuracy for a less complex scene

See also: :class:`Ocr`

- VRAM: 1GB
- Throughput: depends heavily on image resolution, but for 1000x750 images, should be 3-4 images/s

:param paragraph: whether to try to combine boxes together or not
:param resize: whether to resize the images to a reasonable size before sending it over or not. Runs faster if true
:param bs: how many images should this group together and send to the server at once?""" # ocr
        self.paragraph = paragraph; self.resize = resize; self.bs = bs           # ocr
    def __ror__(self, it): return self._all_opt([it]) | cli.item()               # ocr
    def _all_opt(self, it:list["PIL"]):                                          # ocr
        def resize(it): # resizing if they're too big                            # ocr
            for img in it:                                                       # ocr
                w, h = img | cli.shape()                                         # ocr
                if w > h:                                                        # ocr
                    if w > 1000: frac = 1000/w; img = img | tf.Resize([int(h*frac), int(w*frac)]) # ocr
                else:                                                            # ocr
                    if h > 1000: frac = 1000/h; img = img | tf.Resize([int(h*frac), int(w*frac)]) # ocr
                yield img, self.paragraph                                        # ocr
        return (resize(it) if self.resize else it | cli.apply(lambda img: [img, self.paragraph])) | cli.batched(self.bs, True)\
            | cli.apply(lambda imgParas: [imgParas, get("kapi/3-ocr", {"data": k1.encode(imgParas | cli.apply(cli.toBytes(), 0) | cli.deref())}) | cli.aS(k1.decode)] | cli.transpose()) | cli.joinSt() | ~cli.apply(Ocr) # ocr
class Ocr:                                                                       # Ocr
    def __init__(self, imgPara, res):                                            # Ocr
        """Ocr result object. Stores raw results from model in ``.res`` field and has many
more functionalities. Not intended to be instantiated by the end user. Example::

    ocr = someImg | kapi.ocr() # ocr is an object of type Ocr
    ocrBox = ocr[3] # grabs the 3rd detected bounding box

    ocrBox.coords     # grabs coordinates
    ocrBox.text       # grabs recognized text
    ocrBox.confidence # grabs confidence

    ocrBox | toImg()               # grabs image cutout
    ocrBox | toNdArray()           # grabs numpy array cutout
    ocrBox | toImg() | kapi.tess() # returns string, pass the image through tesseract OCR, to get more reliable results

See also: :class:`OcrBox`"""                                                     # Ocr
        self.img, self.para = imgPara; self._npImg = None; self.res = res | cli.apply(~cli.aS(lambda x1,x2,y1,y2: [x1,y1,x2,y2]), 0) | cli.deref() # Ocr
    def npImg(self):                                                             # Ocr
        """Grabs the numpy array of the image, shape (C, H, W)"""                # Ocr
        if self._npImg is None: self._npImg = self.img | cli.toNdArray()         # Ocr
        return self._npImg                                                       # Ocr
    def __repr__(self): return f"<Ocr shape={self.img | cli.shape()}>"           # Ocr
    def _overlay(self) -> "PIL":                                                 # Ocr
        img = self.img; res = self.res; p5 = k1.p5; w, h = img | cli.shape(); p5.newSketch(*img | cli.shape()); p5.background(255); p5.fill(255, 0) # Ocr
        res | cli.cut(0) | ~cli.apply(lambda x1,y1,x2,y2: [x1,h-y2,x2-x1,y2-y1]) | ~cli.apply(p5.rect) | cli.deref() # Ocr
        res | cli.cut(0, 1) | ~cli.apply(lambda x1,y1,x2,y2: [min(x1,x2), h-max(y1,y2)], 0) | ~cli.apply(lambda xy,s: [s,*xy]) | ~cli.apply(p5.text) | cli.deref() # Ocr
        im2 = p5.img(); alpha = 0.3; return [img, im2] | cli.apply(cli.toTensor() | cli.op()[:3]) | ~cli.aS(lambda x,y: x*alpha+y*(1-alpha)) | cli.op().to(int) | cli.op().permute(1, 2, 0) | cli.toImg() # Ocr
    def _repr_html_(self): s = html.escape(f"{self}"); return f"<pre>{s}</pre><img src='data:image/jpeg;base64, {base64.b64encode(self._overlay() | cli.toBytes()).decode()}' />" # Ocr
    def __getitem__(self, s):                                                    # Ocr
        if isinstance(s, slice): return [OcrBox(self, i) for i in range(len(self.res))[s]] # Ocr
        return OcrBox(self, s)                                                   # Ocr
    def __len__(self): return len(self.res)                                      # Ocr
    def __getstate__(self): d = {**self.__dict__}; d["img"] = self.img | cli.toBytes(); d["_npImg"] = None; return d # better compression due to converting to jpg # Ocr
    def __setstate__(self, d): self.__dict__.update(d); self.img = self.img | cli.toImg(); self._npImg = None # Ocr
class OcrBox:                                                                    # OcrBox
    def __init__(self, ocr, i):                                                  # OcrBox
        """1 bounding box of the ocr-ed image. Not intended to be instantiated by the end user.
Example::

    ocr = someImg | kapi.ocr()
    ocrBox = ocr[3] # grabs the 3rd detected bounding box

See also: :class:`Ocr`"""                                                        # OcrBox
        self.ocr = ocr; self.i = i                                               # OcrBox
    @property                                                                    # OcrBox
    def coords(self): return self.ocr.res[self.i][0]                             # OcrBox
    @property                                                                    # OcrBox
    def text(self): return self.ocr.res[self.i][1]                               # OcrBox
    @property                                                                    # OcrBox
    def confidence(self): return self.ocr.res[self.i][2]                         # OcrBox
    def _toNdArray(self):                                                        # OcrBox
        x1,y1,x2,y2 = self.ocr.res[self.i][0]                                    # OcrBox
        return self.ocr.npImg()[:,y1:y2,x1:x2]                                   # OcrBox
    def _toImg(self, **kwargs): return self._toNdArray().transpose((1, 2, 0)) | cli.toImg() # OcrBox
    def __repr__(self): return f"<OcrBox i={self.i} coords={self.coords} confidence={round(self.confidence, 3)} text='{self.text}' />" # OcrBox
    def _repr_html_(self): s = html.escape(f"{self}"); return f"<pre>{s}</pre><img src='data:image/jpeg;base64, {base64.b64encode(self | cli.toImg() | cli.toBytes()).decode()}' />" # OcrBox
init.addAtomic(Ocr); init.addAtomic(OcrBox)                                      # OcrBox
class tess(BaseCli):                                                             # tess
    def __init__(self):                                                          # tess
        """Do OCR using tesseract, instead of easyocr. This is meant for simple images only,
preferably sections cut off from :class:`ocr`. For complex bounding box detection, still
use :class:`ocr`. Example::

    # returns "some text"
    image | kapi.tess()

For small texts, can reach throughput up to 75/s"""                              # tess
        pass                                                                     # tess
    def __ror__(self, it): return self._all_opt([it]) | cli.item()               # tess
    def _all_opt(self, it:list[str]):                                            # tess
        for b in it | cli.batched(75, True):                                     # tess
            yield from get("kapi/12-tess", {"imgs": k1lib.encode(b)})            # tess
class txt2im(BaseCli):                                                           # txt2im
    def __init__(self, num_inference_steps=10):                                  # txt2im
        """Generates images from text descriptions, using stable diffusion v2.
Example::

    "a bowl of apples" | kapi.txt2im() # returns PIL image

- VRAM: 5.42GB
- Throughput: 1/s
"""                                                                              # txt2im
        self.num_inference_steps = num_inference_steps                           # txt2im
    def __ror__(self, it): return get("kapi/4-txt2im", {"prompt": it, "num_inference_steps": self.num_inference_steps}) | cli.aS(base64.b64decode) | cli.toImg() # txt2im
class caption(BaseCli):                                                          # caption
    def __init__(self):                                                          # caption
        """Captions images using model `Salesforce/blip-image-captioning-large`.
Example::

    img = "some_image.png" | toImg() # loads PIL image
    img | kapi.caption()                  # returns string description

- VRAM: 2.5GB
- Throughput: 16/s
"""                                                                              # caption
        pass                                                                     # caption
    def __ror__(self, it): return self._all_opt([it]) | cli.item()               # caption
    def _all_opt(self, it:list["PIL"]): return it | cli.batched(5, True) | cli.apply(lambda imgs: get("kapi/5-caption", {"images": k1lib.encode(imgs)}) | cli.aS(k1lib.decode)) | cli.joinSt() # caption
class speech(BaseCli):                                                           # speech
    def __init__(self, sep=False):                                               # speech
        """Converts English speech to text using whisper-large-v2.
Example::

    "audio.mp3" | toAudio() | kapi.speech() # returns string transcript

- VRAM: 4GB
- Throughput: 20min video finish transcribing in ~25s, so around 60x faster than real time

If the input audio is too long (>25 minutes), then it will be broken up
into multiple smaller pieces around 20 min each and sent to the server,
so at the bounds, it might go wrong a little bit

:param sep: if True, separate transcripts of each segment (returns List[transcript]),
    if False (default), joins segment's transcripts together into a single string""" # speech
        self.sep = sep                                                           # speech
    def __ror__(self, audio:"conv.Audio"):                                       # speech
        nSplits = int(audio.raw.duration_seconds/60/25)+1                        # speech
        res = audio | cli.splitW(*[1]*nSplits) | cli.apply(lambda piece: get("kapi/6-speech", {"audio": base64.b64encode(piece | cli.toBytes()).decode()})) # speech
        return list(res) if self.sep else res | cli.join(". ")                   # speech
def _summarize(text:str) -> str:                                                 # _summarize
    return text | segment(2000) | complete("<|end of transcript|>\n\nPlease summarize the above transcript using 1-3 sentences: ").all()\
        | cli.op().strip().all() | cli.deref() | cli.join(". ")                  # _summarize
class summarize(BaseCli):                                                        # summarize
    def __init__(self, length=1000):                                             # summarize
        """Summarizes text in multiple stages until it's shorter than ``length`` in
characters or until further compression is not possible. Example::

    url = "https://www.youtube.com/watch?v=NfmSjGbnEWk"
    audio = url   | toAudio()     # downloads audio from youtube
    text  = audio | kapi.speech() # does speech recognition
    text | summarize()            # summarize the text. For a 23 minute video/22k characters text, it should take around 23s to summarize everything

This will return an array of strings::

    [
        "shortened text final stage",
        "shortened text stage 2",
        "shortened text stage 1",
        "original text",
    ]

So in each stage, the original text is split up into multiple pieces, then
each piece is summarized using :class:`complete` and then all summary will
be joined together, creating the "shortened text stage 1". This continues
until it the text's length does not decrease any further, or it's shorter
than the desired length.

:param length: desired summary string length"""                                  # summarize
        self.length = length                                                     # summarize
    def __ror__(self, text:str):                                                 # summarize
        stages = [text]; l = len(text)                                           # summarize
        while True:                                                              # summarize
            if len(text) < self.length: return stages | cli.reverse() | cli.deref() # summarize
            l = len(text); text = _summarize(text); stages.append(text)          # summarize
            if len(text)/l > 0.8: return stages | cli.reverse() | cli.deref() # if length not shrinking, then just return early # summarize
class post(BaseCli):                                                             # post
    def __init__(self, url):                                                     # post
        """Creates a post request from a URL that can be read using :meth:`~k1lib.cli.inp.cat`.
Example::

    # returns str of the results
    {"some": "json data"} | kapi.post("https://some.url/some/path")

Notice how there isn't a get request counterpart, because you can always just cat() them
directly, as get requests don't have a body::

    cat("https://some.url/some/path")
"""                                                                              # post
        self.url = url                                                           # post
    def __ror__(self, d): return requests.post(self.url, json=d).text            # post
    def _jsF(self, meta):                                                        # post
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto()                        # post
        return f"""\
const {fIdx} = async ({dataIdx}) => {{
    const res = await fetch({json.dumps(self.url)}, {{ method: "POST", headers: {{ "Content-Type": "application/json" }}, body: JSON.stringify({dataIdx}) }});
    if (res.ok) return await res.text();
    throw new Error(`Can't send POST request to '{self.url}': ${{res.status}} - ${{res.statusText}}`);
}}""", fIdx                                                                      # post



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
__all__ = ["dummy"]
def dummy():                                                                     # dummy
    """Does nothing. Only here for you to checkout the source code"""            # dummy
    pass                                                                         # dummy
from k1lib.imports import *; import hashlib                                      # dummy
def hashF(msg:str) -> str: m = hashlib.sha256(); m.update(f"{msg}".encode()); return k1.encode(m.digest()) # hashF
def cpuHash() -> str: return None | cmd("lscpu") | head() | join("\n") | aS(hashF) # cpuHash
loadTestFn = "~/.k1lib/applyCl_loadTest.pth"                                     # cpuHash
def load_loadTest(): return cat(loadTestFn, False) | aS(dill.loads) if os.path.exists(os.path.expanduser(loadTestFn)) else dict() # load_loadTest
def good_loadTest(): # whether the underlying architecture has been load-tested before # good_loadTest
    obj = load_loadTest(); return None | applyCl.aS(lambda: cpuHash()) | cut(1) | ~inSet(obj) | shape(0) == 0 # good_loadTest
def loadTestS(nodeId, cpus, hash_):                                              # loadTestS
    with k1.timer() as t1:                                                       # loadTestS
        [nodeId]*cpus*4 | insertIdColumn(begin=False) | applyCl(lambda x: range(300_000_000) | toSum(), pre=True) | deref() # loadTestS
    with k1.timer() as t2:                                                       # loadTestS
        [nodeId]*cpus*4 | insertIdColumn(begin=False) | applyCl(lambda x: range(30_000_000) | apply(op()+2) | toSum(), pre=True) | deref() # loadTestS
    return [nodeId, cpus, hash_, t1(), t2()]                                     # loadTestS
def loadTest():                                                                  # loadTest
    data = None | applyCl.aS(lambda: [os.cpu_count(), cpuHash()] | deref()) | ~apply(lambda x,y: [x,*y]) | deref() # loadTest
    a = data | ~applyTh(loadTestS, timeout=3600) | deref()                       # loadTest
    alpha1, alpha2 = a | cut(1, 3, 4) | ~apply(lambda x,y,z: [y, z]) | transpose() | toMax().all() | deref() # loadTest
    obj = a | ~apply(lambda idx,cpu,h,t1,t2: [h, (alpha1/(t1), alpha2/(t2)) | toMean()]) | toDict() | deref() # loadTest
    obj | aS(dill.dumps) | file(loadTestFn)                                      # loadTest
def loadTestGuard(guard=True, rounded=True): # returns Dict[nodeId, cpus]        # loadTestGuard
    if guard and not good_loadTest():                                            # loadTestGuard
        ans = input("""applyCl has not load-tested your system yet, so running
the requested operation might be unbalanced on the cluster (some
nodes finish before others, wasting computational time). Would you
like to perform a load test now? Should take 1-2 minutes. Y/n: """)              # loadTestGuard
        if ans.lower()[0] == "y": loadTest()                                     # loadTestGuard
    data = None | applyCl.aS(lambda: [cpuHash(), os.cpu_count()]) | apply(wrapList(), 0) | joinStreams().all() | deref() # List[nodeId, cpu hash, #cpus] # loadTestGuard
    obj = {**data | cut(1, 2) | apply("1", 1) | toDict(), **load_loadTest()}     # loadTestGuard
    return data | lookup(obj, 1) | ~apply(lambda nodeId,mul,cpu: [nodeId, mul*cpu | (aS(round) if rounded else iden())]) | toDict() # loadTestGuard
def balancedNodeIds(): a = loadTestGuard().items(); return a | ~apply(lambda x,y: [x]*y) | joinStreams() | randomize(None) | repeatFrom() | randomize(a | cut(1) | toSum() | op()*2) # balancedNodeIds
from k1lib.imports import *                                                      # balancedNodeIds
getFolderSize = ls() | filt(os.path.isdir).split() | apply(lambda x: x | (tryout(0) | getFolderSize)) + apply(os.path.getsize) | toSum().all() | toSum() | deref() # balancedNodeIds
getFilesInFolder = aS(os.path.expanduser) | aS(os.walk) | cut(0, 2) | ungroup() | join(os.sep).all() # balancedNodeIds
def getIr(base): return None | applyCl.aS(lambda: ls(base) | iden() & apply(lambda x: x | (tryout(0) | (aS(os.path.getsize) if os.path.isfile(x) else getFolderSize))) | transpose() | deref()) | ungroup(False) | insertIdColumn(True) | deref() # getIr
_normF1 = cut(1) | toSum()                                                       # getIr
def normalize(d):                                                                # normalize
    s = 0                                                                        # normalize
    for a,b in d: s+=b                                                           # normalize
    return [[a,b/s] for a,b in d]                                                # normalize
    return d | apply(op()/s, 1) | deref()                                        # normalize
@lru_cache                                                                       # normalize
def statsCpu(nodeIds):                                                           # statsCpu
    cpu = loadTestGuard(False).items() | inSet(nodeIds, 0) | sort(0, False) | deref() # statsCpu
    cpuF = loadTestGuard(False).items() | inSet(nodeIds, 0) | sort(0, False) | deref() | aS(normalize); cpuF # "cpuF" = cpu fraction. List[nodeId, cpu fraction] # statsCpu
    return [cpu, cpuF]                                                           # statsCpu
@lru_cache                                                                       # statsCpu
def statsNodeId(): return applyCl.nodeIds()                                      # statsNodeId
_statsS1 = statsNodeId() | apply(wrapList() | insert(0, False)) | toDict()       # statsNodeId
def stats(inter, nodeIds): # inter for cached data                               # stats
    sizeF = normalize(inter["sizes"].items()) # size fraction. List[nodeId, size fraction] # stats
    return *statsCpu(nodeIds), sizeF                                             # stats
_scoresF = cut(1).all() | transpose() | ~apply(lambda c,s: s-c) | deref()        # stats
def scores(inter, nodeIds): # 0 for best situation, negative means more compute than size, positive means more size than compute # scores
    cpu, cpuF, sizeF = stats(inter, nodeIds)                                     # scores
    return [cpuF, sizeF] | _scoresF # negpos                                     # scores
scoreF = apply(lambda x: abs(x)**2) | toSum()                                    # scores
def getIrp(ir, nodeIds): # "processed ir". Dict[nodeId -> Dict[idx -> Tuple[url, size]]]. This will mutate across optimizations and is intended to be fast # getIrp
    return [nodeIds | apply("[x, {}]"), ir | groupBy(1, True) | apply(~apply(lambda idx,url,size: [idx,[url,size]]) | toDict(), 1)] | joinStreams() | sort(0, False) | toDict() # getIrp
def getSizes(irp): return irp.items() | apply(op().values() | cut(1) | toSum(), 1) | toDict() # Dict[nodeId -> int] for total size of a node # getSizes
def move(ir, nA:str, nB:str, idx:int, inter=None): # mutates ir                  # move
    ir1 = ir; ir1[idx][1] = nB                                                   # move
    if inter is not None:                                                        # move
        irp = inter["irp"]; irp[nB][idx] = elem = irp[nA][idx]; del irp[nA][idx] # move
        sizes = inter["sizes"]; sizes[nB] += elem[1]; sizes[nA] -= elem[1]       # move
    return ir1                                                                   # move
fileSizesF = apply(lambda arr: np.array([e[1] for e in arr[1].values()]))        # move
def optimize(ir, nodeIds, inter=None): # inter for intermediary data, to speed things up. Has "irp", "sizes" # optimize
    hasInter = inter is not None; inter = inter or {}                            # optimize
    irp = inter["irp"] if hasInter else getIrp(ir, nodeIds)                      # optimize
    sizes = inter["sizes"] if hasInter else getSizes(irp)                        # optimize
    cpu, cpuF = statsCpu(nodeIds); inter = {**inter, "irp": irp, "sizes": sizes} # optimize
    scs = inter["scs"] if hasInter else scores(inter, nodeIds)                   # optimize
    a = np.argmax(scs); b = np.argmin(scs)                                       # optimize
    files = list(irp.items()); fA = files[a]; fB = files[b] # files A. Tuple[nodeId, Dict[idx -> Tuple[url, size]]]. Previously Tuple[nodeId, List[idx, url, size]] # optimize
    nA = nodeIds[a]; nB = nodeIds[b] # A node id, str                            # optimize
    sA, sB = [fA, fB] | fileSizesF # file sizes in A, List[int]                  # optimize
    # print(f"sA: {sA} {sB}")                                                    # optimize
    spA = sA.sum() - sA; spB = sB.sum() + sA # sum prime A, array[files]         # optimize
    sp = spA + spB; sfA = spA/sp; sfB = spB/sp                                   # optimize
    cA = cpu[a][1]; cB = cpu[b][1] # cpu A                                       # optimize
    c = cA + cB; cfA = cA/c; cfB = cB/c # cpu fraction A                         # optimize
    # print(f"sfA: {sfA}, cfA: {cfA}")                                           # optimize
    exp = 5 # intuition says that exp should be even. But that doesn't work. Odd values work tho, but I have no idea why # optimize
    i = ((sfA-cfA)**exp + (sfB-cfB)**exp).argmin()                               # optimize
    idx = fA[1].keys() | rItem(i)                                                # optimize
    ir2 = move(ir, nA, nB, idx, inter)                                           # optimize
    scs = scores(inter, nodeIds); inter["scs"] = scs                             # optimize
    return ir2, [nA, nB, idx, scs | scoreF], inter                               # optimize
def traj(ir, maxSteps=20, nodeIds=None):                                         # traj
    nodeIds = nodeIds | sort(None, False) | deref() | aS(tuple); irp = getIrp(ir, nodeIds); sizes = getSizes(irp) # traj
    sc = scores({"irp": irp, "sizes": sizes}, nodeIds) | scoreF; aux = None; inter = None; auxs = []; irs = [] # traj
    maxSteps = maxSteps if maxSteps is not None else int(1e10)                   # traj
    for i in range(maxSteps):                                                    # traj
        print(f"\rVirtually moving files and folders around. Step {i} of total max steps ({maxSteps})...", end="") # traj
        ir, aux, inter = optimize(ir, nodeIds, inter)                            # traj
        if aux[3] > sc: break                                                    # traj
        irs.append(ir); auxs.append(aux); sc = aux[3]                            # traj
    print(); return irs, auxs                                                    # traj
def collapse(it):                                                                # collapse
    a, b = it | rows(0, -1); c = [a[0], b[1], a[2], b[3]]                        # collapse
    return [] if c[0] == c[1] else [c]                                           # collapse
def traj2(ir, traj): # just looks up the file names really, no processing involved # traj2
    idx2FileName = ir | apply(lambda arr: [arr[0], arr[2]]) | toDict()           # traj2
    a = traj | groupBy(2) | filt(lambda x: len(x) > 1).split() | (apply(collapse)) + iden() | joinStreams(2) | deref() # traj2
    return a | lookup(idx2FileName, 2) | deref()                                 # traj2
# def moveFile(fileName:str, sourceNodeId:str, destNodeId:str, timeout=60): # old, slow, corrupted version # traj2
#     """Moves file from the current node to the destination node. Usually executed on other nodes than the driver node""" # traj2
#     fn = os.path.expanduser(fileName); dirname = os.path.dirname(fn)           # traj2
#     applyCl.cmd(f"mkdir -p {dirname}", nodeIds=[destNodeId]); applyCl.cmd(f"rm -f {fn}", nodeIds=[destNodeId]) # traj2
#     for chunk in cat(fn, False, True): [destNodeId] | applyCl.aS(lambda: chunk >> file(fn), timeout=timeout) | deref() # traj2
#     None | cmd(f"rm {fn}") | deref(); return "ok1"                             # traj2
def moveFile(fn:str, sourceN:str, destN:str, timeout=None): # runs on dest node  # moveFile
    fn = os.path.expanduser(fn); dirname = os.path.dirname(fn)                   # moveFile
    applyCl.cmd(f"mkdir -p {dirname}; rm -f {fn}", nodeIds=[destN]);             # moveFile
    windows = [sourceN] | applyCl.aS(lambda: range(os.path.getsize(fn)) | batched(settings.cli.cat.chunkSize, True) | apply("[x.start, x.stop]") | deref(), num_cpus=0.2, timeout=timeout) | cut(1) | item() | deref() # moveFile
    for chunk in [[sourceN]*len(windows), windows] | transpose() | ~applyCl(lambda sB,eB: cat(fn, False, sB=sB, eB=eB), pre=True, prefetch=20, num_cpus=0.2, timeout=timeout) | cut(1): chunk >> file(fn) # moveFile
    applyCl.cmd(f"rm {fn}", nodeIds=[sourceN])                                   # moveFile
def moveFF(ff:str, sourceN:str, destN:str, timeout=None): # runs on dest node    # moveFF
    """Moves file or folder from the current node to the destination node"""     # moveFF
    ff = os.path.expanduser(ff); isfile = [sourceN] | applyCl.aS(lambda: os.path.isfile(ff)) | cut(1) | item() # moveFF
    if isfile: return moveFile(ff, sourceN, destN, timeout)                      # moveFF
    [sourceN] | applyCl.aS(lambda: ff | getFilesInFolder | deref(), timeout=timeout) | cut(1) | item() | apply(aS(moveFile, sourceN, destN, timeout)) | deref() # moveFF
    applyCl.cmd(f"rm -rf {ff}", nodeIds=[sourceN])                               # moveFF
def moveAll(tr, bs=5, timeout=None):                                             # moveAll
    groups = tr | groupBy(1) | deref() # grouping by destination                 # moveAll
    with ray.progress(len(groups), "Moving files around") as rp:                 # moveAll
        def process(idx, arrs): # processing requests for dest node              # moveAll
            arrss = arrs | batched(3, True) | deref()                            # moveAll
            for i, arrs in enumerate(arrss): # moveFF(fn, a, b)                  # moveAll
                arrs | apply(lambda arr: [arr[1], arr]) | ~applyCl(lambda a,b,fn,sc: moveFF(fn, a, b, timeout), pre=True, timeout=timeout) | deref() # move 3 files on dest node # moveAll
                rp.update.remote(idx, (i+1)/len(arrss))                          # moveAll
        groups | insertIdColumn() | ~applyTh(process, timeout=1e9) | deref()     # moveAll
    # tr | apply(lambda arr: [arr[0], arr]) | ~applyCl(lambda a,b,fn,sc: moveFF(fn, b), pre=True, timeout=timeout) | deref() # old version without progress bar # moveAll
def balanceFolder(base, audit=False, maxSteps=1000, timeout=None, bs=5): # currently executing each move step serially, will change in the future if it's too slow # balanceFolder
    loadTestGuard(); applyCl.cmd(f"mkdir -p {base}"); ir = getIr(base)           # balanceFolder
    tr = traj2(ir, traj(ir, maxSteps, tuple(applyCl.nodeIds()))[1]); return tr if audit else moveAll(tr, bs, timeout) # balanceFolder
def decommissionFolderTraj(base:str, nAs:List[str]): # nAs are the ones to decommission # decommissionFolderTraj
    ir = getIr(base); targetNodes = applyCl.nodeIds() | ~inSet(nAs) | repeatFrom(); irs = []; auxs = [] # decommissionFolderTraj
    seis = ir | inSet(nAs, 1) | cut(0, 1) | ~apply(lambda idx, startNode: [startNode, next(targetNodes), idx, 0]) | deref() # List[start, end, index, score] # decommissionFolderTraj
    for sei in seis: ir = move(ir, *sei[:3]); irs.append(ir); auxs.append(sei)   # decommissionFolderTraj
    return irs, auxs                                                             # decommissionFolderTraj
def decommissionFolder(base:str, nAs:List[str], audit=False, maxSteps=1000, timeout=None, bs=5): # decommissionFolder
    loadTestGuard(); irs, auxs = decommissionFolderTraj(base, nAs)               # decommissionFolder
    nBs=applyCl.nodeIds() | ~inSet(nAs) | aS(tuple)                              # decommissionFolder
    if len(nBs) > 1:                                                             # decommissionFolder
        irs2, auxs2 = traj(irs[-1] if len(irs) > 0 else getIr(base), maxSteps, nBs) # decommissionFolder
        irs = [*irs, *irs2]; auxs = [*auxs, *auxs2]                              # decommissionFolder
    if len(irs) == 0: return                                                     # decommissionFolder
    tr = traj2(irs[-1], auxs)                                                    # decommissionFolder
    if audit: return tr                                                          # decommissionFolder
    moveAll(tr, bs, timeout); nAs | applyCl.aS(lambda: None | cmd(f"rm -rf {base}") | deref()) | deref() # move all files and then deletes the empty folders # decommissionFolder
def a_transfer(fn, nse, nodeB, rpF:callable=iden()):                             # a_transfer
    """Transfers a lot of blocks from a bunch of nodes to nodeB. Does not delete from those node though

nse = List[nodeAId, [sB, eB]]

Runs on driver process, blocks, so better use applyTh outside of this

:param rpF: ray progress function"""                                             # a_transfer
    blockSize = settings.cli.cat.chunkSize                                       # a_transfer
    def inner():                                                                 # a_transfer
        totalBytes = nse | cut(1) | ~apply(lambda x,y:y-x) | toSum(); currentByte = 0 # a_transfer
        for chunk in nse | ~apply(lambda x, y: range(x, y) | batched(blockSize, True) | apply("[x.start, x.stop]"), 1) | ungroup() | deref()\
            | ~applyCl(lambda sB, eB: cat(fn, False, sB=sB, eB=eB), pre=True, timeout=None, prefetch=20) | cut(1): # a_transfer
            chunk >> file(fn); currentByte += len(chunk); rpF(currentByte/totalBytes) # a_transfer
    [nodeB] | applyCl.aS(inner, timeout=None) | item()                           # a_transfer
def decommission(fn:str, nodeAs:List[str], nodeBs:List[str], rS):                # decommission
    """Spreads out a particular file in nodeAs to all nodeBs, to prepare
to decomission nodeAs. The 2 sets should be mutually exclusive

:param rS: instance of refineSeek"""                                             # decommission
    nodeAs, nodeBs = [nodeAs, nodeBs] | deref()                                  # decommission
    if len(nodeAs) == 0: return                                                  # decommission
    if len(nodeBs) == 0: raise Exception("Unsupported configuration! Trying to move data from A+B to C+D. Has to have some shared nodes, like moving data from A+B+C to B+C+D. This is not a fundamental limitation, but just can't be done with the current architecture. Might be fixed in the future.") # decommission
    # some initial metadata                                                      # decommission
    nodeIds = applyCl.nodeIds(); nodeId_cpu = loadTestGuard(False).items() | deref(); nodeId2Cpu = nodeId_cpu | toDict() # decommission
    ws = nodeId_cpu | inSet(nodeBs, 0) | cut(1) | deref() # weights to split files on nodeAs into # decommission
    # splitting file on nodeAs into chunks first, to plan things out             # decommission
    a = nodeAs | applyCl.aS(lambda: fn | splitSeek(ws=ws) | rS | window(2) | deref() | insertColumn(nodeBs) | insert(applyCl.nodeId()).all() | deref()) | cut(1) | joinStreams() | deref() # decommission
    # actually transferring chunks                                               # decommission
    with ray.progress(a | groupBy(1) | shape(0), "Decommissioning") as rp:       # decommission
        c = b = a | groupBy(1, True) | apply(iden() + apply(lambda arr: [arr[0], arr[1:]]) | reverse() | insert(fn)) | deref() # decommission
        enumerate(c) | applyTh(~aS(lambda idx, e: a_transfer(*e, rpF=aS(lambda p: ray.get(rp.update.remote(idx, p))))), timeout=1e9) | deref() # decommission
    # deleting files from nodeAs                                                 # decommission
    nodeAs | applyCl.aS(lambda: None | cmd(f"rm -rf {fn}") | ignore()) | deref() # decommission
ranges2Seeks = apply("[x.start, x.stop]") | joinStreams() | aS(set) | sort(None) | deref() # decommission
def spreadOut(fn:str, nAs:List[str], nBs:List[str], rS):                         # spreadOut
    """Spreads out a file from nodes A to B, where B fully contains A (no decomissioning).
A and B should be mutually exclusive. Initial nodes are A, final nodes are A + B""" # spreadOut
    nAs, nBs = [nAs, nBs] | deref(); rS.fn = fn                                  # spreadOut
    if len(nBs) == 0: return # no need to spread out                             # spreadOut
    nBs | applyCl.aS(lambda: None | cmd(f"mkdir -p {os.path.dirname(fn)}") | deref(), timeout=None) | deref() # spreadOut
    nBs | applyCl.aS(lambda: None | cmd(f"rm -rf {fn}") | deref(), timeout=None) | deref() # spreadOut
    # some initial metadata                                                      # spreadOut
    nodeIds = applyCl.nodeIds(); nodeId_cpu = loadTestGuard(False).items() | deref(); nodeId2Cpu = nodeId_cpu | toDict() # spreadOut
    sizes = nAs | applyCl.aS(lambda: os.path.getsize(fn) if os.path.exists(fn) else 0) | deref(); totalSize = sizes | cut(1) | toSum() # spreadOut
    ns = [*nAs, *nBs]; totalCpu = ns | lookup(nodeId2Cpu) | toSum(); bytePerCpu = totalSize/totalCpu; wsB = nBs | lookup(nodeId2Cpu) | deref() # spreadOut
    # prepares segments and metadata, List[nodeId, [sB, eB]], where sB and eB are the ranges of nAs that they're willing to share # spreadOut
    sizePost = sizes | ~apply(lambda idx, size: [idx, nodeId2Cpu[idx]/totalCpu*totalSize/size]) | deref() # size fraction to retain # spreadOut
    invalidNodes = sizePost | ~filt(lambda x: 0 <= x <= 1, 1) | cut(0) | deref() # spreadOut
    if len(invalidNodes) > 0: raise Exception(f"Unsupported configuration! These nodes have too little data to share: {invalidNodes}. This couldn't have happen using applyCl alone. Data is not corrupted, but you'll have to combine data from all files into 1 and spread them back out again.") # spreadOut
    inter = sizePost | ~apply(lambda idx, x: [idx, [x, 1-x]]) | applyCl(lambda ws: fn | splitSeek(ws=ws) | rS | ~head(1), pre=True, timeout=None) | deref() | filt(~aS(lambda x,y: y-x>0), 1) | deref() # filter at the end to eliminate files that don't want to share anything (x == y) # spreadOut
    # actually transferring data to new nodes                                    # spreadOut
    meta = inter | apply(~aS(range) | splitW(*wsB) | ranges2Seeks | apply(lambda x: splitSeek.backward(fn, x)) | deref() | rS | window(2) | deref() | apply(wrapList()) | insertColumn(nBs), 1) | ungroup(False) | groupBy(1, True) | deref() # spreadOut
    with ray.progress(len(meta), "Transferring data to new nodes") as rp:        # spreadOut
        meta | insertIdColumn(True) | applyTh(~aS(lambda idx, nB, nse: a_transfer(fn, nse, nB, rpF=aS(lambda p: ray.get(rp.update.remote(idx, p))))), timeout=1e9) | deref() # spreadOut
    # truncates the files in nAs nodes                                           # spreadOut
    inter | ~apply(lambda idx,se: [idx,se[0]]) | applyCl(lambda sB: open(fn, 'a').truncate(sB), pre=True, timeout=None) | deref() # spreadOut
def balanceFile(fn:str, nAs:List[str]=None, nBs:List[str]=None, rS=None):        # balanceFile
    fn = os.path.expanduser(fn); rS = rS or refineSeek(); rS.injectFn(fn); loadTestGuard() # balanceFile
    if nAs is None: nAs = None | applyCl.aS(lambda: os.path.exists(fn)) | filt(op(), 1) | cut(0) | deref() # balanceFile
    if nBs is None: nBs = applyCl.nodeIds()                                      # balanceFile
    decommission(fn, *nAs | inSet(nBs).split() | reverse(), rS)                  # balanceFile
    spreadOut(fn, *nBs | inSet(nAs).split(), rS)                                 # balanceFile
def diskScan1(base:str) -> List[str]: # like ls(), but returns files and folders that appear at least on 2 nodes. No recursion # diskScan1
    isdir, base = base.split("\ue000")                                           # diskScan1
    if not isdir: return []                                                      # diskScan1
    return None | applyCl.aS(lambda: base | (tryout([]) | ls() | apply(os.path.isdir) & iden() | transpose() | ~apply(lambda x,y: f"{x*1}\ue000{y}")) | deref(), timeout=60) | cut(1) | joinStreams() | count() | filt(op()>1, 0) | cut(1) | deref() # diskScan1
def diskScan2(base:str) -> Tuple[List[str], List[str]]: # returns list of distributed folders and list of distributed files # diskScan2
    dFolders = []; folders, files = diskScan1(base) | op().split("\ue000").all() | toInt(0) | filt(op(), 0).split() | (join("\ue000")).all(2) | deref() # first explore this directory # diskScan2
    # print("2--", folders, files, base)                                         # diskScan2
    for folder in folders: # then recursively explore inside directories         # diskScan2
        fol, fil = diskScan2(folder); dFolders.extend(fol); files.extend(fil)    # diskScan2
        if len(fol) + len(fil) == 0: dFolders.append(folder) # no shared contents, must be a distributed folder # diskScan2
        else: files.extend(fil)                                                  # diskScan2
    # print("3--", [dFolders, files], base)                                      # diskScan2
    return [dFolders, files]                                                     # diskScan2
def getFolderSize2(folder:str):                                                  # getFolderSize2
    folder = os.path.expanduser(folder)                                          # getFolderSize2
    return None | cmd(f"du -s {folder}") | table() | cut(0) | item() | aS(int) | op()*1024 # getFolderSize2
def diskScan3(base:str, accurate=False):                                         # diskScan3
    base = os.path.expanduser(base); folders, files = diskScan2(f"1\ue000{base}") | op().split("\ue000")[1].all(2) | apply(set) | apply(list) | aS(list) # getting rid of all file/folder flags # diskScan3
    getFZ = getFolderSize if accurate else getFolderSize2                        # diskScan3
    folders = [folders, None | applyCl.aS(lambda: folders | apply(lambda x: getFZ(x)           if os.path.exists(x) else 0) | deref(), timeout=300) | cut(1) | transpose()] | transpose() | deref() # diskScan3
    files   = [files,   None | applyCl.aS(lambda: files   | apply(lambda x: os.path.getsize(x) if os.path.exists(x) else 0) | deref(), timeout=300) | cut(1) | transpose()] | transpose() | deref() # diskScan3
    # this section below tries to squeeze out replicatedFolders from a bunch of replicatedFiles. The exact mechanism involved may seem like magic to you, but it seems to work # diskScan3
    balancedFolders = folders; replicatedFiles, balancedFiles = files | filt(filt("x") | aS(set) | shape(0) | (op() == 1), 1).split() | deref() # diskScan3
    f1 = iden() & apply(os.path.dirname) | joinStreams() | aS(set)               # diskScan3
    excludedFolders = balancedFolders | cut(0) | serial(*[f1]*100) | aS(set)     # diskScan3
    f = iden() & apply(os.path.dirname) | joinStreams() | ~inSet(excludedFolders); elims = []; i = 0 # diskScan3
    a = replicatedFiles | cut(0) | apply(os.path.dirname) | serial(*[f]*30) | aS(set) | sort(None, False) | aS(list) # diskScan3
    while i < len(a): # trying to eliminate child directories, so that replicatedFolders work recursively. Doing this weird index loop so that time complexity is O(n*log(n)) instead of O(n^2) # diskScan3
        j = i+1                                                                  # diskScan3
        while j < len(a) and a[j].startswith(a[i]): elims.append(a[j]); j += 1   # diskScan3
        i = j                                                                    # diskScan3
    candidates = a | ~inSet(elims) | deref() # replicatedFolder candidates. Has to reverify that total folder size is the same before declaring it a replicated folder. Yes, this is not the same as each individual files are the same, and I can imagine an edge case that will trip this up, but it's so damn rare and I'm so lazy that I'm sticking with this version # diskScan3
    replicatedFolders = [candidates, None | applyCl.aS(lambda: candidates | apply(lambda x: getFolderSize(x) if os.path.exists(x) else 0) | deref()) | cut(1) | transpose()] | transpose() | apply(filt("x") | aS(set) | shape(0) | aS("x == 1"), 1) | filt("x", 1) | cut(0) | deref() # diskScan3
    replicatedFiles = replicatedFiles | ~filt(lambda fn: replicatedFolders | filt(lambda folder: fn.startswith(f"{folder}/")) | shape(0), 0) | deref() # diskScan3
    replicatedFolders = [replicatedFolders, None | applyCl.aS(lambda: replicatedFolders | apply(lambda x: getFZ(x) if os.path.exists(x) else 0) | deref()) | cut(1) | transpose()] | transpose() | deref() # diskScan3
    return balancedFolders, replicatedFolders, balancedFiles, replicatedFiles    # diskScan3
def diskScan4(base:str, sortSize=True, accurate=False): # fully featured data    # diskScan4
    return diskScan3(base, accurate) | (apply(~sortF(toSum(), 1)) if sortSize else iden()) | deref() # diskScan4
def diskScan5(base:str, sortSize=True, accurate=False, f=iden()): # displays it in a nice format # diskScan5
    d4 = diskScan4(base, sortSize, accurate) | f; lens = d4 | apply(len) | deref(); nodeNames = None | applyCl.aS(lambda: os.cpu_count()) | apply(op()[:5], 0) | apply('f"{x} thr"', 1) | join(", ").all() | deref(); nodeNames # diskScan5
    d5 = d4 | apply(~apply(lambda path, sizes: [path, sizes | toSum() | aS(fmt.size), sizes | apply(fmt.size)]) | insert(["-"*40, "-"*10, ["-"*12]*len(nodeNames)]) | insert(["", "", nodeNames])) | deref(); d5 # diskScan5
    ws = d5 | shape(0).all() | deref()                                           # diskScan5
    d6 = d5 | joinStreams() | cut(0, 1) & (cut(2) | pretty() | wrapList().all()) | transpose() | joinStreams().all() | splitW(*ws) | insert(["Path", "Total size", "Size on each node (node id and thread count)"]).all() | joinStreams() | pretty() | splitW(*ws | apply(op()+1)) | deref() # diskScan5
    explainers = ["\nA distributed folder is a folder that has many files and folders inside, but their names\nare all different from each other. It's managed by applyCl.balanceFolder()", # diskScan5
                  "\nA replicated folder is a folder that has been copied to multiple nodes. Size of all folder\ncopies should be the same. It's managed by applyCl.replicateFolder()", # diskScan5
                  "\nA distributed file is a file that has been split into multiple pieces and sent to other\nnodes. It's managed by applyCl.balanceFile()", # diskScan5
                  "\nA replicated file is a file that has been copied to multiple nodes. Size of all file\ncopies should be the same. It's managed by applyCl.replicateFile()",] # diskScan5
    arr = [d6, ["Distributed folders", "Replicated folders", "Distributed files", "Replicated files"] | (aS(lambda x: [["="*60, x, "="*60] | join(" ")])).all()] | transpose() | permute(1, 0) | (joinStreams() | join("\n")).all() | wrapList() | insert(explainers, False) | transpose() | join("\n").all() | deref() # diskScan5
    [arr, lens] | transpose() | filt(op(), 1) | cut(0) | join("\n"*2) | wrapList() | stdout() # diskScan5
